==> go-test-framework-231.9161.41-sources.jar: com/goide/GoTestUtil.java <==
// Copyright 2016-2022 JetBrains s.r.o.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package com.goide;

import com.goide.codeInsight.imports.GoImportsSettings;
import com.goide.documentation.GoDocumentationProvider;
import com.goide.inspections.core.GoInspectionBase;
import com.goide.project.*;
import com.goide.psi.GoFile;
import com.goide.psi.GoType;
import com.goide.refactor.util.GoRefactoringUtil;
import com.goide.sdk.GoSdk;
import com.goide.sdk.GoSdkService;
import com.goide.sdk.GoSdkUtil;
import com.goide.vgo.configuration.VgoProjectSettings;
import com.intellij.analysis.AnalysisScope;
import com.intellij.codeHighlighting.HighlightDisplayLevel;
import com.intellij.codeInsight.daemon.HighlightDisplayKey;
import com.intellij.codeInsight.documentation.DocumentationManager;
import com.intellij.codeInsight.generation.surroundWith.SurroundWithHandler;
import com.intellij.codeInspection.ex.InspectionProfileImpl;
import com.intellij.codeInspection.ex.InspectionToolWrapper;
import com.intellij.ide.DataManager;
import com.intellij.lang.documentation.DocumentationProvider;
import com.intellij.openapi.Disposable;
import com.intellij.openapi.actionSystem.*;
import com.intellij.openapi.application.PathManager;
import com.intellij.openapi.application.WriteAction;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.fileEditor.FileEditor;
import com.intellij.openapi.fileEditor.FileEditorManager;
import com.intellij.openapi.fileEditor.TextEditor;
import com.intellij.openapi.module.Module;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.roots.FileIndexFacade;
import com.intellij.openapi.util.Disposer;
import com.intellij.openapi.util.io.FileUtil;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.openapi.vfs.LocalFileSystem;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.vfs.newvfs.impl.VfsRootAccess;
import com.intellij.profile.codeInspection.InspectionProfileManager;
import com.intellij.profile.codeInspection.ProjectInspectionProfileManager;
import com.intellij.psi.PsiDocumentManager;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import com.intellij.psi.search.GlobalSearchScope;
import com.intellij.psi.search.ProjectScopeImpl;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.testFramework.PlatformTestUtil;
import com.intellij.testFramework.TestModeFlags;
import com.intellij.testFramework.UsefulTestCase;
import com.intellij.testFramework.fixtures.CodeInsightTestFixture;
import com.intellij.util.ObjectUtils;
import com.intellij.util.ReflectionUtil;
import com.intellij.util.SmartList;
import com.intellij.util.ThreeState;
import junit.framework.TestCase;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.File;
import java.io.IOException;
import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;
import java.util.Objects;

import static com.goide.GoTestUtilKt.runBlockingAction;
import static com.intellij.testFramework.UsefulTestCase.assertSize;

public final class GoTestUtil {
  @NotNull
  public static File getTestDataPath(@NotNull String path) {
    String homePath = PathManager.getHomePath();
    File testData = FileUtil.findFirstThatExist(homePath + "/goland/intellij-go/impl/testData/" + path, // go tests in ide
                                                homePath + "/impl/testData/" + path, // ???
                                                homePath + "/goland/intellij-go/" + path, // vgo tests in ide
                                                "impl/testData/" + path, // go tests
                                                path); // relative to working dir or absolute
    return Objects.requireNonNull(testData);
  }

  public static AnAction findSurroundAction(@NotNull String surrounderDescription, @NotNull PsiFile file, @NotNull Editor editor) {
    List<AnAction> applicable = SurroundWithHandler.buildSurroundActions(file.getProject(), editor, file, null);
    if (applicable == null) {
      return null;
    }
    String suffix = ". " + surrounderDescription;
    for (AnAction action : applicable) {
      String actionPresentation = action.getTemplatePresentation().getText();
      if (actionPresentation != null && StringUtil.endsWith(actionPresentation, suffix)) {
        return action;
      }
    }
    return null;
  }

  public static void setupGoSupport(@NotNull UsefulTestCase testCase,
                                    @NotNull Project project,
                                    @NotNull Module module,
                                    boolean addTmpDirToLibraryRoots,
                                    @Nullable String defaultSdkVersion) {
    GoImportsSettings.getInstance(project).setOptimizeImportsOnTheFly(true);
    GoProjectLibrariesService.getInstance(project).setIndexEntireGopath(true);
    GoProjectLibrariesService.getInstance(project).setUseGoPathFromSystemEnvironment(false);
    GoModuleSettings.getInstance(module).setGoSupportEnabled(true);
    GoModuleSettings.getInstance(module).setVendoringEnabled(ThreeState.YES);
    GoModuleSettings.getInstance(module).setBuildTargetSettings(new GoBuildTargetSettings());
    SdkAware sdkAwareAnnotation = findAnnotation(testCase, SdkAware.class);
    boolean needSDK = sdkAwareAnnotation != null && !"NULL".equals(sdkAwareAnnotation.value());
    if (needSDK) {
      setUpProjectSdk(testCase, project);
      String sdkCustomVersion = sdkAwareAnnotation.value();
      if (!sdkCustomVersion.isEmpty()) {
        TestModeFlags.set(GoSdkService.TESTING_SDK_VERSION, sdkCustomVersion, testCase.getTestRootDisposable());
      }
      else if (StringUtil.isNotEmpty(defaultSdkVersion)) {
        TestModeFlags.set(GoSdkService.TESTING_SDK_VERSION, defaultSdkVersion, testCase.getTestRootDisposable());
      }
    }
    else {
      GoSdkService.getInstance(project).setSdk(GoSdk.NULL);
    }
    List<String> libraryRootUrls = new SmartList<>();
    if (addTmpDirToLibraryRoots) {
      libraryRootUrls.add("temp:///");
    }
    if (needSDK) {
      libraryRootUrls.add(setUpMockGoPath(testCase).getUrl());
    }
    GoApplicationLibrariesService.getInstance().setLibraryRootUrls(libraryRootUrls);
    GoProjectLibrariesService.getInstance(project).setLibraryRootUrls();
    GoModuleLibrariesService.getInstance(module).setLibraryRootUrls();
    VgoProjectSettings.getInstance(project).setIntegrationEnabled(false);
    PlatformTestUtil.dispatchAllInvocationEventsInIdeEventQueue();
  }

  private static void setUpProjectSdk(@NotNull UsefulTestCase testCase, @NotNull Project project) {
    String homePath = FileUtil.toCanonicalPath(getTestDataPath("mockSdk/").getAbsolutePath());
    VfsRootAccess.allowRootAccess(testCase.getTestRootDisposable(), homePath);
    GoSdkService.getInstance(project).setSdk(GoSdk.fromHomePath(homePath));
  }

  private static VirtualFile setUpMockGoPath(@NotNull UsefulTestCase testCase) {
    String homePath = FileUtil.toCanonicalPath(getTestDataPath("mockGoPath/").getAbsolutePath());
    VirtualFile gopath = LocalFileSystem.getInstance().findFileByPath(homePath);
    VfsRootAccess.allowRootAccess(testCase.getTestRootDisposable(), homePath);
    return Objects.requireNonNull(gopath);
  }

  public static <T extends Annotation> T findAnnotation(@NotNull UsefulTestCase testCase, @NotNull Class<T> annotationClass) {
    Class<?> aClass = testCase.getClass();
    String methodName = "test" + UsefulTestCase.getTestName(testCase.getName(), false);
    boolean methodChecked = false;
    T annotation;
    while (aClass != null && aClass != Object.class) {
      if (!methodChecked) {
        Method method = ReflectionUtil.getDeclaredMethod(aClass, methodName);
        if (method != null) {
          annotation = method.getAnnotation(annotationClass);
          if (annotation != null) return annotation;
          methodChecked = true;
        }
      }
      annotation = aClass.getAnnotation(annotationClass);
      if (annotation != null) return annotation;
      aClass = aClass.getSuperclass();
    }
    return null;
  }

  public static <T extends GoInspectionBase> @NotNull T findInspection(@NotNull Project project, @NotNull Class<T> clazz) {
    InspectionProfileImpl inspectionProfile = InspectionProfileManager.getInstance(project).getCurrentProfile();
    try {
      InspectionToolWrapper<?, ?> tool = inspectionProfile.getInspectionTool(clazz.getConstructor().newInstance().getShortName(), project);
      T inspection = ObjectUtils.tryCast(tool != null ? tool.getTool() : null, clazz);
      if (inspection != null) return inspection;
      throw new IllegalStateException("failed to find inspection " + clazz.getSimpleName());
    }
    catch (InstantiationException | IllegalAccessException | InvocationTargetException | NoSuchMethodException e) {
      throw new IllegalStateException("failed to find inspection " + clazz.getSimpleName(), e);
    }
  }

  @NotNull
  public static AnalysisScope getGoFilesScope(@NotNull Project project) {
    FileIndexFacade index = FileIndexFacade.getInstance(project);
    GlobalSearchScope searchScope = GlobalSearchScope.getScopeRestrictedByFileTypes(
      new ProjectScopeImpl(project, index) {
        @Override
        public boolean contains(@NotNull VirtualFile file) {
          return super.contains(file) && !index.isInLibrarySource(file);
        }
      }, GoFileType.INSTANCE
    );
    AnalysisScope scope = new AnalysisScope(searchScope, project);
    scope.invalidate();
    return scope;
  }

  public static void copyTestDataToProject(@NotNull CodeInsightTestFixture fixture, @NotNull String testDataDir, @NotNull String targetDir) {
    fixture.getTempDirFixture().copyAll(getTestDataPath(testDataDir).getAbsolutePath(), targetDir);
  }

  public static void setSeverity(@NotNull Class<? extends GoInspectionBase> inspection,
                                 @NotNull HighlightDisplayLevel severity,
                                 @NotNull Project project) {
    String inspectionName = findInspection(project, inspection).getShortName();
    ProjectInspectionProfileManager.getInstance(project).getCurrentProfile()
      .setErrorLevel(HighlightDisplayKey.find(inspectionName), severity, project);
  }

  @SuppressWarnings("ConstantConditions")
  public static @NotNull VirtualFile findSdkFile(@NotNull CodeInsightTestFixture fixture, @NotNull String path) {
    return GoSdkService.getInstance(fixture.getProject()).getSdk(fixture.getModule()).getSrcDir().findFileByRelativePath(path);
  }

  @NotNull
  public static VirtualFile findGopathFile(@NotNull CodeInsightTestFixture fixture, @NotNull String path) {
    return GoSdkUtil.getGoPathSources(fixture.getProject(), fixture.getModule(), true).stream()
      .map(src -> src.findFileByRelativePath(path))
      .filter(Objects::nonNull)
      .findFirst()
      .orElseThrow(() -> new IllegalArgumentException("Cannot file file in GOPATH: " + path));
  }

  public static void newScratchFile(@NotNull Editor editor) {
    AnAction action = ActionManager.getInstance().getAction("NewScratchFile");

    DataContext context = DataManager.getInstance().getDataContext(editor.getComponent());
    AnActionEvent e = AnActionEvent.createFromDataContext(ActionPlaces.EDITOR_TAB, null, context);

    action.actionPerformed(e);
  }

  public static @Nullable VirtualFile getScratchVirtualFile(@NotNull Project project, @NotNull Disposable disposable) {
    FileEditor[] editors = FileEditorManager.getInstance(project).getAllEditors();
    assertSize(2, editors);

    for (FileEditor fileEditor : editors) {
      Editor editor = ((TextEditor)fileEditor).getEditor();
      Document document = editor.getDocument();
      PsiFile file = PsiDocumentManager.getInstance(project).getPsiFile(document);
      if (file != null && file.getName().startsWith("scratch")) {
        VirtualFile virtualFile = file.getVirtualFile();
        Disposer.register(disposable, () -> {
          try {
            WriteAction.run(() -> virtualFile.delete(disposable));
          }
          catch (IOException e) {
            throw new RuntimeException(e);
          }
        });
        return virtualFile;
      }
    }

    return null;
  }

  public static void checkDocumentationMarkup(@NotNull CodeInsightTestFixture fixture, @NotNull String expectedMarkupFile) {
    Editor editor = fixture.getEditor();
    PsiFile file = fixture.getFile();

    var elementAndContext =
      DocumentationManager.getInstance(fixture.getProject()).findTargetElementAndContext(editor, editor.getCaretModel().getOffset(), file);
    TestCase.assertNotNull(elementAndContext);
    PsiElement docElement = elementAndContext.first;
    PsiElement originalElement = elementAndContext.second;
    TestCase.assertNotNull(originalElement);
    DocumentationProvider documentationProvider = DocumentationManager.getProviderFromElement(originalElement);

    String actualDoc = runBlockingAction(() -> StringUtil.notNullize(documentationProvider.generateDoc(docElement, originalElement)));
    String quickNavigateInfo = runBlockingAction(() -> documentationProvider.getQuickNavigateInfo(docElement, originalElement));

    actualDoc += "\n=====\n" + StringUtil.notNullize(quickNavigateInfo, "No navigation info");

    List<String> urls = documentationProvider.getUrlFor(docElement, originalElement);
    actualDoc += "\n=====\n" + (urls != null ? StringUtil.join(urls, "\n") : "No urls");

    PsiElement adjustedDocElement = GoDocumentationProvider.adjustDocElement(docElement);
    TestCase.assertNotNull(adjustedDocElement);
    String localUrl = GoDocumentationProvider.getLocalUrlToElement(adjustedDocElement);
    actualDoc += "\n=====\n" + StringUtil.notNullize(localUrl, "No local urls");
    UsefulTestCase.assertSameLinesWithFile(expectedMarkupFile, actualDoc);

    if (localUrl != null) {
      PsiElement elementForLink =
        documentationProvider.getDocumentationElementForLink(fixture.getPsiManager(), localUrl, originalElement);
      TestCase.assertNotNull("Link '" + localUrl + "' wasn't resolved", elementForLink);
      TestCase.assertEquals("Link was resolved to unexpected element.\nExpected: " + docElement.getText() + "\n" +
                            "Actual: " + elementForLink.getText(), adjustedDocElement, elementForLink);
    }
  }

  public static void doGetTypeTextWithImportsTest(@NotNull GoFile context, @NotNull GoFile fileWithType, @NotNull String expectedTypeText) {
    GoType type = PsiTreeUtil.findChildOfType(fileWithType, GoType.class);
    TestCase.assertNotNull(type);
    GoRefactoringUtil.TypeTextWithImports typeTextWithImports = GoRefactoringUtil.getTypeTextWithImports(context, type);
    TestCase.assertEquals(expectedTypeText, typeTextWithImports.typeText);
  }
}

==> go-test-framework-232.10072.27-sources.jar: com/goide/GoTestUtil.java <==
// Copyright 2016-2022 JetBrains s.r.o.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package com.goide;

import com.goide.codeInsight.imports.GoImportsSettings;
import com.goide.documentation.GoDocumentationProvider;
import com.goide.inspections.core.GoInspectionBase;
import com.goide.project.*;
import com.goide.psi.GoFile;
import com.goide.psi.GoType;
import com.goide.refactor.util.GoRefactoringUtil;
import com.goide.sdk.GoSdk;
import com.goide.sdk.GoSdkService;
import com.goide.sdk.GoSdkUtil;
import com.goide.sdk.GoTargetSdkVersionProvider;
import com.goide.vgo.configuration.VgoProjectSettings;
import com.intellij.analysis.AnalysisScope;
import com.intellij.codeHighlighting.HighlightDisplayLevel;
import com.intellij.codeInsight.daemon.HighlightDisplayKey;
import com.intellij.codeInsight.documentation.DocumentationManager;
import com.intellij.codeInsight.generation.surroundWith.SurroundWithHandler;
import com.intellij.codeInspection.ex.InspectionProfileImpl;
import com.intellij.codeInspection.ex.InspectionToolWrapper;
import com.intellij.ide.DataManager;
import com.intellij.lang.documentation.DocumentationProvider;
import com.intellij.openapi.Disposable;
import com.intellij.openapi.actionSystem.*;
import com.intellij.openapi.application.PathManager;
import com.intellij.openapi.application.WriteAction;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.fileEditor.FileEditor;
import com.intellij.openapi.fileEditor.FileEditorManager;
import com.intellij.openapi.fileEditor.TextEditor;
import com.intellij.openapi.module.Module;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.roots.FileIndexFacade;
import com.intellij.openapi.util.Disposer;
import com.intellij.openapi.util.io.FileUtil;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.openapi.vfs.LocalFileSystem;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.vfs.newvfs.impl.VfsRootAccess;
import com.intellij.profile.codeInspection.InspectionProfileManager;
import com.intellij.profile.codeInspection.ProjectInspectionProfileManager;
import com.intellij.psi.PsiDocumentManager;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import com.intellij.psi.search.GlobalSearchScope;
import com.intellij.psi.search.ProjectScopeImpl;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.testFramework.PlatformTestUtil;
import com.intellij.testFramework.TestModeFlags;
import com.intellij.testFramework.UsefulTestCase;
import com.intellij.testFramework.fixtures.CodeInsightTestFixture;
import com.intellij.util.ObjectUtils;
import com.intellij.util.ReflectionUtil;
import com.intellij.util.SmartList;
import junit.framework.TestCase;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.File;
import java.io.IOException;
import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;
import java.util.Objects;

import static com.goide.GoTestUtilKt.runBlockingAction;
import static com.intellij.testFramework.UsefulTestCase.assertSize;

public final class GoTestUtil {
  @NotNull
  public static File getTestDataPath(@NotNull String path) {
    String homePath = PathManager.getHomePath();
    File testData = FileUtil.findFirstThatExist(homePath + "/goland/intellij-go/impl/testData/" + path, // go tests in ide
                                                homePath + "/impl/testData/" + path, // ???
                                                homePath + "/goland/intellij-go/" + path, // vgo tests in ide
                                                "impl/testData/" + path, // go tests
                                                path); // relative to working dir or absolute
    return Objects.requireNonNull(testData);
  }

  public static AnAction findSurroundAction(@NotNull String surrounderDescription, @NotNull PsiFile file, @NotNull Editor editor) {
    List<AnAction> applicable = SurroundWithHandler.buildSurroundActions(file.getProject(), editor, file, null);
    if (applicable == null) {
      return null;
    }
    String suffix = ". " + surrounderDescription;
    for (AnAction action : applicable) {
      String actionPresentation = action.getTemplatePresentation().getText();
      if (actionPresentation != null && StringUtil.endsWith(actionPresentation, suffix)) {
        return action;
      }
    }
    return null;
  }

  public static void setupGoSupport(@NotNull UsefulTestCase testCase,
                                    @NotNull Project project,
                                    @NotNull Module module,
                                    boolean addTmpDirToLibraryRoots,
                                    @Nullable String defaultSdkVersion) {
    GoImportsSettings.getInstance(project).setOptimizeImportsOnTheFly(true);
    GoProjectLibrariesService.getInstance(project).setIndexEntireGopath(true);
    GoProjectLibrariesService.getInstance(project).setUseGoPathFromSystemEnvironment(false);
    GoModuleSettings.getInstance(module).setGoSupportEnabled(true);
    GoModuleSettings.getInstance(module).setBuildTargetSettings(new GoBuildTargetSettings());
    SdkAware sdkAwareAnnotation = findAnnotation(testCase, SdkAware.class);
    boolean needSDK = sdkAwareAnnotation != null && !"NULL".equals(sdkAwareAnnotation.value());
    if (needSDK) {
      setUpProjectSdk(testCase, project);
      String sdkVersion = sdkAwareAnnotation.value().isEmpty() ? defaultSdkVersion : sdkAwareAnnotation.value();
      TestModeFlags.set(GoSdkService.TESTING_SDK_VERSION, sdkVersion, testCase.getTestRootDisposable());
      setTargetGoVersion(testCase, sdkAwareAnnotation, sdkVersion);
    }
    else {
      GoSdkService.getInstance(project).setSdk(GoSdk.NULL);
    }
    List<String> libraryRootUrls = new SmartList<>();
    if (addTmpDirToLibraryRoots) {
      libraryRootUrls.add("temp:///");
    }
    if (needSDK) {
      libraryRootUrls.add(setUpMockGoPath(testCase).getUrl());
    }
    GoApplicationLibrariesService.getInstance().setLibraryRootUrls(libraryRootUrls);
    GoProjectLibrariesService.getInstance(project).setLibraryRootUrls();
    GoModuleLibrariesService.getInstance(module).setLibraryRootUrls();
    VgoProjectSettings.getInstance(project).setIntegrationEnabled(false);
    PlatformTestUtil.dispatchAllInvocationEventsInIdeEventQueue();
  }

  private static void setTargetGoVersion(@NotNull UsefulTestCase testCase, @NotNull SdkAware sdkAwareAnnotation, @Nullable String sdkVersion) {
    var targetVersion = switch (sdkAwareAnnotation.targetVersion()) {
      case "" -> sdkVersion;
      case "null" -> null;
      default -> sdkAwareAnnotation.targetVersion();
    };
    setTargetGoVersion(targetVersion, testCase.getTestRootDisposable());
  }

  private static void setTargetGoVersion(@Nullable String targetGoVersion, @NotNull Disposable disposable) {
    TestGoTargetSdkVersionProvider provider = new TestGoTargetSdkVersionProvider(targetGoVersion);
    GoTargetSdkVersionProvider.EP_NAME.getPoint().registerExtension(provider, disposable);
  }

  private static class TestGoTargetSdkVersionProvider implements GoTargetSdkVersionProvider {
    private final String myTargetGoVersion;

    private TestGoTargetSdkVersionProvider(@Nullable String targetGoVersion) {
      myTargetGoVersion = targetGoVersion;
    }

    @Override
    public boolean isApplicable(@NotNull PsiFile file) {
      return true;
    }

    @Override
    public @Nullable String getTargetSdkVersion(@NotNull PsiFile file) {
      return myTargetGoVersion;
    }
  }

  private static void setUpProjectSdk(@NotNull UsefulTestCase testCase, @NotNull Project project) {
    String homePath = FileUtil.toCanonicalPath(getTestDataPath("mockSdk/").getAbsolutePath());
    VfsRootAccess.allowRootAccess(testCase.getTestRootDisposable(), homePath);
    GoSdkService.getInstance(project).setSdk(GoSdk.fromHomePath(homePath));
  }

  private static VirtualFile setUpMockGoPath(@NotNull UsefulTestCase testCase) {
    String homePath = FileUtil.toCanonicalPath(getTestDataPath("mockGoPath/").getAbsolutePath());
    VirtualFile gopath = LocalFileSystem.getInstance().findFileByPath(homePath);
    VfsRootAccess.allowRootAccess(testCase.getTestRootDisposable(), homePath);
    return Objects.requireNonNull(gopath);
  }

  public static <T extends Annotation> T findAnnotation(@NotNull UsefulTestCase testCase, @NotNull Class<T> annotationClass) {
    Class<?> aClass = testCase.getClass();
    String methodName = "test" + UsefulTestCase.getTestName(testCase.getName(), false);
    boolean methodChecked = false;
    T annotation;
    while (aClass != null && aClass != Object.class) {
      if (!methodChecked) {
        Method method = ReflectionUtil.getDeclaredMethod(aClass, methodName);
        if (method != null) {
          annotation = method.getAnnotation(annotationClass);
          if (annotation != null) return annotation;
          methodChecked = true;
        }
      }
      annotation = aClass.getAnnotation(annotationClass);
      if (annotation != null) return annotation;
      aClass = aClass.getSuperclass();
    }
    return null;
  }

  public static <T extends GoInspectionBase> @NotNull T findInspection(@NotNull Project project, @NotNull Class<T> clazz) {
    InspectionProfileImpl inspectionProfile = InspectionProfileManager.getInstance(project).getCurrentProfile();
    try {
      InspectionToolWrapper<?, ?> tool = inspectionProfile.getInspectionTool(clazz.getConstructor().newInstance().getShortName(), project);
      T inspection = ObjectUtils.tryCast(tool != null ? tool.getTool() : null, clazz);
      if (inspection != null) return inspection;
      throw new IllegalStateException("failed to find inspection " + clazz.getSimpleName());
    }
    catch (InstantiationException | IllegalAccessException | InvocationTargetException | NoSuchMethodException e) {
      throw new IllegalStateException("failed to find inspection " + clazz.getSimpleName(), e);
    }
  }

  @NotNull
  public static AnalysisScope getGoFilesScope(@NotNull Project project) {
    FileIndexFacade index = FileIndexFacade.getInstance(project);
    GlobalSearchScope searchScope = GlobalSearchScope.getScopeRestrictedByFileTypes(
      new ProjectScopeImpl(project, index) {
        @Override
        public boolean contains(@NotNull VirtualFile file) {
          return super.contains(file) && !index.isInLibrarySource(file);
        }
      }, GoFileType.INSTANCE
    );
    AnalysisScope scope = new AnalysisScope(searchScope, project);
    scope.invalidate();
    return scope;
  }

  public static void copyTestDataToProject(@NotNull CodeInsightTestFixture fixture, @NotNull String testDataDir, @NotNull String targetDir) {
    fixture.getTempDirFixture().copyAll(getTestDataPath(testDataDir).getAbsolutePath(), targetDir);
  }

  public static void setSeverity(@NotNull Class<? extends GoInspectionBase> inspection,
                                 @NotNull HighlightDisplayLevel severity,
                                 @NotNull Project project) {
    String inspectionName = findInspection(project, inspection).getShortName();
    ProjectInspectionProfileManager.getInstance(project).getCurrentProfile()
      .setErrorLevel(HighlightDisplayKey.find(inspectionName), severity, project);
  }

  @SuppressWarnings("ConstantConditions")
  public static @NotNull VirtualFile findSdkFile(@NotNull CodeInsightTestFixture fixture, @NotNull String path) {
    return GoSdkService.getInstance(fixture.getProject()).getSdk(fixture.getModule()).getSrcDir().findFileByRelativePath(path);
  }

  @NotNull
  public static VirtualFile findGopathFile(@NotNull CodeInsightTestFixture fixture, @NotNull String path) {
    return GoSdkUtil.getGoPathSources(fixture.getProject(), fixture.getModule(), true).stream()
      .map(src -> src.findFileByRelativePath(path))
      .filter(Objects::nonNull)
      .findFirst()
      .orElseThrow(() -> new IllegalArgumentException("Cannot file file in GOPATH: " + path));
  }

  public static void newScratchFile(@NotNull Editor editor) {
    AnAction action = ActionManager.getInstance().getAction("NewScratchFile");

    DataContext context = DataManager.getInstance().getDataContext(editor.getComponent());
    AnActionEvent e = AnActionEvent.createFromDataContext(ActionPlaces.EDITOR_TAB, null, context);

    action.actionPerformed(e);
  }

  public static @Nullable VirtualFile getScratchVirtualFile(@NotNull Project project, @NotNull Disposable disposable) {
    FileEditor[] editors = FileEditorManager.getInstance(project).getAllEditors();
    assertSize(2, editors);

    for (FileEditor fileEditor : editors) {
      Editor editor = ((TextEditor)fileEditor).getEditor();
      Document document = editor.getDocument();
      PsiFile file = PsiDocumentManager.getInstance(project).getPsiFile(document);
      if (file != null && file.getName().startsWith("scratch")) {
        VirtualFile virtualFile = file.getVirtualFile();
        Disposer.register(disposable, () -> {
          try {
            WriteAction.run(() -> virtualFile.delete(disposable));
          }
          catch (IOException e) {
            throw new RuntimeException(e);
          }
        });
        return virtualFile;
      }
    }

    return null;
  }

  public static void checkDocumentationMarkup(@NotNull CodeInsightTestFixture fixture, @NotNull String expectedMarkupFile) {
    Editor editor = fixture.getEditor();
    PsiFile file = fixture.getFile();

    var elementAndContext =
      DocumentationManager.getInstance(fixture.getProject()).findTargetElementAndContext(editor, editor.getCaretModel().getOffset(), file);
    TestCase.assertNotNull(elementAndContext);
    PsiElement docElement = elementAndContext.first;
    PsiElement originalElement = elementAndContext.second;
    TestCase.assertNotNull(originalElement);
    DocumentationProvider documentationProvider = DocumentationManager.getProviderFromElement(originalElement);

    String actualDoc = runBlockingAction(() -> StringUtil.notNullize(documentationProvider.generateDoc(docElement, originalElement)));
    String quickNavigateInfo = runBlockingAction(() -> documentationProvider.getQuickNavigateInfo(docElement, originalElement));

    actualDoc += "\n=====\n" + StringUtil.notNullize(quickNavigateInfo, "No navigation info");

    List<String> urls = documentationProvider.getUrlFor(docElement, originalElement);
    actualDoc += "\n=====\n" + (urls != null ? StringUtil.join(urls, "\n") : "No urls");

    PsiElement adjustedDocElement = GoDocumentationProvider.adjustDocElement(docElement);
    TestCase.assertNotNull(adjustedDocElement);
    String localUrl = GoDocumentationProvider.getLocalUrlToElement(adjustedDocElement);
    actualDoc += "\n=====\n" + StringUtil.notNullize(localUrl, "No local urls");
    UsefulTestCase.assertSameLinesWithFile(expectedMarkupFile, actualDoc);

    if (localUrl != null) {
      PsiElement elementForLink =
        documentationProvider.getDocumentationElementForLink(fixture.getPsiManager(), localUrl, originalElement);
      TestCase.assertNotNull("Link '" + localUrl + "' wasn't resolved", elementForLink);
      TestCase.assertEquals("Link was resolved to unexpected element.\nExpected: " + docElement.getText() + "\n" +
                            "Actual: " + elementForLink.getText(), adjustedDocElement, elementForLink);
    }
  }

  public static void doGetTypeTextWithImportsTest(@NotNull GoFile context, @NotNull GoFile fileWithType, @NotNull String expectedTypeText) {
    GoType type = PsiTreeUtil.findChildOfType(fileWithType, GoType.class);
    TestCase.assertNotNull(type);
    GoRefactoringUtil.TypeTextWithImports typeTextWithImports = GoRefactoringUtil.getTypeTextWithImports(context, type);
    TestCase.assertEquals(expectedTypeText, typeTextWithImports.typeText);
  }
}
